#include <stdlib.h>
#include <stdio.h>
#include <string.h>

enum errors {
  ERR_NO_EQ = -0x001,

  ERR_ALLOC_OPT = -0x010,
  ERR_ALLOC_STR = -0x020,
  ERR_OPEN_FILE = -0x040,
};

struct option {
  char *var;
  char *value;
  struct option *next;
};

/* Returns the position of the '=' char in a string. Useful for splitting it in
 * two. find_eq_pos will return ERR_NO_EQ if no '=' is found so you can ignore
 * it or handle it differently. */
int find_eq_position(const char *line);

/* Returns a string without comments */
void trim_comments(char *line);

/* Returns a trimmed string (with no trailing or leading spaces). Eg
 * 'option=value' == '  option  = value '. It's not very good, memorywise, but
 * it works. */
char *trim_white_space(char *line);

/* Adds a new node at the start of the list. Returns 0 on success or
 * ERR_ALLOC_OPT if there were a problem when calling malloc. */
int new_option(struct option **options, char *var, char *value);

/* Reverses the list. Useful because as we push() things, they will be in the
 * opposite order from the config file, so if we want to access it in the same
 * way, we should reverse it. Returns the reversed list. */
struct option *reverse(struct option *begin);

/* Parses a line and, if successful, add it to the options list. It may return
 * ERR_ALLOC_OPT or ERR_ALLOC_STR, which says there were a problem calling
 * malloc, or 0, success */
int parse_line(struct option **options, char *line);

/* Searches for the option named opt_name. If none is found, returns NULL, if
 * finds, returns the list node whose var name is opt_name. This is useful as
 * there may be more than one option with the same opt_name, so you can search
 * again, running get_opt with return->next. Don't forget, however, that it
 * returns the original node; changes directly made directly to it may corrupt
 * your opts list. */
struct option *get_option(struct option *options, const char *option_name);

/* Parses the file. */
int parse_file(struct option **options, const char *file_name);

int main(int argc, char *argv[])
{
  struct option *options = NULL;
  struct option *current = NULL;

  if(argc == 1)
  {
    parse_file(&options, "file-config.cfg");
  } else {
    parse_file(&options, argv[1]);
  }

  current = options;
  while(current != NULL)
  {
    printf("%s: %s\n", current->var, current->value);
    current = current->next;
  }

  current = options;
  while( (current = get_option(current, "message")) ) /* yes, it's a =, not a == */
  {
    printf("%s found: %s\n", current->var, current->value);
    current = current->next;
  }

  return(0);
}

int find_eq_position(const char *line)
{
  int count = 0;

  while(*(line+count) != '')
  {
    if(*(line+count) == '=')
    {
      return(count);
    }
    count++;
  }

  return(ERR_NO_EQ);
}

void trim_comments(char *line)
{
  int count = 0;

  while(*(line+count) != '')
  {
    if(*(line+count) == '#')
    {
      *(line+count) = '';
      return;
    }

    count++;
  }
}

char *trim_white_space(char *line)
{
  char *end = NULL;

  while(*line == ' ' || *line == '\t' || *line == '\n' || *line == '\r')
  {
    line++;
  }

  end = line + strlen(line) - 1;
  while(end > line &&
                (*end == ' ' || *end == '\t' || *end == '\n' || *end == '\r'))
  {
    end--;
  }

  *(end+1) = '';
  return(line);
}

int new_option(struct option **options, char *var, char *value)
{
  struct option *new = malloc(sizeof(struct option));
  if(new == NULL)
  {
#ifdef DEBUG
    fprintf(stderr, "error allocating %u bytes of memory in 'new_option'\n",
            sizeof(struct option));
#endif
    return(ERR_ALLOC_OPT);
  }

  new->var = var;
  new->value = value;
  new->next = *options;

  *options = new;

  return(0);
}

struct option *reverse(struct option *begin)
{
  struct option *prev = NULL;
  struct option *curr = begin;
  struct option *next = NULL;

  /* if begin is null, there is no next (segfault) */
  if(begin==NULL)
  {
    return(NULL);
  }
  next = curr->next;

  /* if next is null, this is a single list. return as it is. */
  if(next==NULL)
  {
    return(curr);
  }

  /* otherwise, reverse it */
  while(next!=NULL)
  {
    next = curr->next;
    curr->next = prev;
    prev = curr;
    curr = next;
  }

  return(prev);
}

int parse_line(struct option **options, char *line)
{
  char *var = NULL;
  char *value = NULL;
  char *linecpy = NULL;

  int position = 0;
  int line_size = 0;

  trim_comments(line);

  position = find_eq_position(line);
  if(position == ERR_NO_EQ)
  {
    return(ERR_NO_EQ);
  }

  line_size = strlen(line);
  linecpy = calloc(line_size, sizeof(char)*line_size);
  if(linecpy==NULL)
  {
#ifdef DEBUG
    fprintf(stderr, "error allocating %u bytes of memory in 'parse_line'\n",
            sizeof(char)*line_size);
#endif
    return(ERR_ALLOC_STR);
  }

  strcpy(linecpy, line);

  *(linecpy+position) = '';

  var = trim_white_space(linecpy);
  value = trim_white_space(linecpy + position + 1);

  return(new_option(options, var, value));
}

struct option *get_option(struct option *options, const char *option_name)
{
  struct option *current = options;

  while(current != NULL)
  {
    if(strcmp(current->var, option_name) == 0)
    {
      return(current);
    }
    current = current->next;
  }

  return(NULL);
}

int parse_file(struct option **options, const char *file_name)
{
  char line[256];
  FILE *fptr = NULL;

  fptr = fopen(file_name, "r");
  if(fptr == NULL)
  {
#ifdef DEBUG
    fprintf(stderr, "error openning file %s in 'parse_file'\n", file_name);
#endif
    return(ERR_OPEN_FILE);
  }

  while(fgets(line, 255, fptr) != NULL)
  {
    int ret = parse_line(options, line);
    if(ret == ERR_ALLOC_OPT || ret == ERR_ALLOC_STR)
    {
#ifdef DEBUG
      fprintf(stderr, "error allocating memory in 'parse_file' (%x)\n", ret);
#endif
      fclose(fptr);
      return(ret);
    }
  }

  *options = reverse(*options);

  fclose(fptr);
  return(0);
}
